# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# API Refererence:
#   - https://developer.pagerduty.com/api-reference/\
#       368ae3d938c9e-send-an-event-to-pager-duty
#

from json import dumps

import requests

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool, validate_regex
from .base import NotifyBase


class PagerDutySeverity:
    """Defines the Pager Duty Severity Levels."""

    INFO = "info"

    WARNING = "warning"

    ERROR = "error"

    CRITICAL = "critical"


# Map all support Apprise Categories with the Pager Duty ones
PAGERDUTY_SEVERITY_MAP = {
    NotifyType.INFO: PagerDutySeverity.INFO,
    NotifyType.SUCCESS: PagerDutySeverity.INFO,
    NotifyType.WARNING: PagerDutySeverity.WARNING,
    NotifyType.FAILURE: PagerDutySeverity.CRITICAL,
}

PAGERDUTY_SEVERITIES = (
    PagerDutySeverity.INFO,
    PagerDutySeverity.WARNING,
    PagerDutySeverity.CRITICAL,
    PagerDutySeverity.ERROR,
)


# Priorities
class PagerDutyRegion:
    US = "us"
    EU = "eu"


# SparkPost APIs
PAGERDUTY_API_LOOKUP = {
    PagerDutyRegion.US: "https://events.pagerduty.com/v2/enqueue",
    PagerDutyRegion.EU: "https://events.eu.pagerduty.com/v2/enqueue",
}

# A List of our regions we can use for verification
PAGERDUTY_REGIONS = (
    PagerDutyRegion.US,
    PagerDutyRegion.EU,
)


class NotifyPagerDuty(NotifyBase):
    """A wrapper for Pager Duty Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Pager Duty"

    # The services URL
    service_url = "https://pagerduty.com/"

    # Secure Protocol
    secure_protocol = "pagerduty"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_pagerduty"

    # We don't support titles for Pager Duty notifications
    title_maxlen = 0

    # Allows the user to specify the NotifyImageSize object; this is supported
    # through the webhook
    image_size = NotifyImageSize.XY_128

    # Our event action type
    event_action = "trigger"

    # The default region to use if one isn't otherwise specified
    default_region = PagerDutyRegion.US

    # Define object templates
    templates = (
        "{schema}://{integrationkey}@{apikey}",
        "{schema}://{integrationkey}@{apikey}/{source}",
        "{schema}://{integrationkey}@{apikey}/{source}/{component}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            # Optional but triggers V2 API
            "integrationkey": {
                "name": _("Integration Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "source": {
                # Optional Source Identifier (preferably a FQDN)
                "name": _("Source"),
                "type": "string",
                "default": "Apprise",
            },
            "component": {
                # Optional Component Identifier
                "name": _("Component"),
                "type": "string",
                "default": "Notification",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "group": {
                "name": _("Group"),
                "type": "string",
            },
            "class": {
                "name": _("Class"),
                "type": "string",
                "map_to": "class_id",
            },
            "click": {
                "name": _("Click"),
                "type": "string",
            },
            "region": {
                "name": _("Region Name"),
                "type": "choice:string",
                "values": PAGERDUTY_REGIONS,
                "default": PagerDutyRegion.US,
                "map_to": "region_name",
            },
            # The severity is automatically determined, however you can
            # optionally over-ride its value and force it to be what you want
            "severity": {
                "name": _("Severity"),
                "type": "choice:string",
                "values": PAGERDUTY_SEVERITIES,
                "map_to": "severity",
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": True,
                "map_to": "include_image",
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "details": {
            "name": _("Custom Details"),
            "prefix": "+",
        },
    }

    def __init__(
        self,
        apikey,
        integrationkey=None,
        source=None,
        component=None,
        group=None,
        class_id=None,
        include_image=True,
        click=None,
        details=None,
        region_name=None,
        severity=None,
        **kwargs,
    ):
        """Initialize Pager Duty Object."""
        super().__init__(**kwargs)

        # Long-Lived Access token (generated from User Profile)
        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = f"An invalid Pager Duty API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.integration_key = validate_regex(integrationkey)
        if not self.integration_key:
            msg = (
                "An invalid Pager Duty Routing Key "
                f"({integrationkey}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # An Optional Source
        self.source = self.template_tokens["source"]["default"]
        if source:
            self.source = validate_regex(source)
            if not self.source:
                msg = (
                    "An invalid Pager Duty Notification Source "
                    f"({source}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.component = self.template_tokens["source"]["default"]

        # An Optional Component
        self.component = self.template_tokens["component"]["default"]
        if component:
            self.component = validate_regex(component)
            if not self.component:
                msg = (
                    "An invalid Pager Duty Notification Component "
                    f"({component}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.component = self.template_tokens["component"]["default"]

        # Store our region
        try:
            self.region_name = (
                self.default_region
                if region_name is None
                else region_name.lower()
            )

            if self.region_name not in PAGERDUTY_REGIONS:
                # allow the outer except to handle this common response
                raise IndexError()

        except (AttributeError, IndexError, TypeError):
            # Invalid region specified
            msg = f"The PagerDuty region specified ({region_name}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg) from None

        # The severity (if specified)
        self.severity = (
            None
            if severity is None
            else next(
                (
                    s
                    for s in PAGERDUTY_SEVERITIES
                    if str(s).lower().startswith(severity)
                ),
                False,
            )
        )

        if self.severity is False:
            # Invalid severity specified
            msg = f"The PagerDuty severity specified ({severity}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # A clickthrough option for notifications
        self.click = click

        # Store Class ID if specified
        self.class_id = class_id

        # Store Group if specified
        self.group = group

        self.details = {}
        if details:
            # Store our extra details
            self.details.update(details)

        # Display our Apprise Image
        self.include_image = include_image

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Send our PagerDuty Notification."""

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Authorization": f"Token token={self.apikey}",
        }

        # Prepare our persistent_notification.create payload
        payload = {
            # Define our integration key
            "routing_key": self.integration_key,
            # Prepare our payload
            "payload": {
                "summary": body,
                # Set our severity
                "severity": (
                    PAGERDUTY_SEVERITY_MAP[notify_type]
                    if not self.severity
                    else self.severity
                ),
                # Our Alerting Source/Component
                "source": self.source,
                "component": self.component,
            },
            "client": self.app_id,
            # Our Event Action
            "event_action": self.event_action,
        }

        if self.group:
            payload["payload"]["group"] = self.group

        if self.class_id:
            payload["payload"]["class"] = self.class_id

        if self.click:
            payload["links"] = [{
                "href": self.click,
            }]

        # Acquire our image url if configured to do so
        image_url = (
            None if not self.include_image else self.image_url(notify_type)
        )

        if image_url:
            payload["images"] = [{
                "src": image_url,
                "alt": notify_type,
            }]

        if self.details:
            payload["payload"]["custom_details"] = {}
            # Apply any provided custom details
            for k, v in self.details.items():
                payload["payload"]["custom_details"][k] = v

        # Prepare our URL based on region
        notify_url = PAGERDUTY_API_LOOKUP[self.region_name]

        self.logger.debug(
            "Pager Duty POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Pager Duty Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                notify_url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code not in (
                requests.codes.ok,
                requests.codes.created,
                requests.codes.accepted,
            ):
                # We had a problem
                status_str = NotifyPagerDuty.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Pager Duty notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Pager Duty notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Pager Duty "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.integration_key,
            self.apikey,
            self.source,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "region": self.region_name,
            "image": "yes" if self.include_image else "no",
        }
        if self.class_id:
            params["class"] = self.class_id

        if self.group:
            params["group"] = self.group

        if self.click is not None:
            params["click"] = self.click

        if self.severity:
            params["severity"] = self.severity

        # Append our custom entries our parameters
        params.update({f"+{k}": v for k, v in self.details.items()})

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        url = (
            "{schema}://{integration_key}@{apikey}/"
            "{source}/{component}?{params}"
        )

        return url.format(
            schema=self.secure_protocol,
            # never encode hostname since we're expecting it to be a valid one
            integration_key=self.pprint(
                self.integration_key, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            apikey=self.pprint(
                self.apikey, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            source=self.pprint(self.source, privacy, safe=""),
            component=self.pprint(self.component, privacy, safe=""),
            params=NotifyPagerDuty.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The 'apikey' makes it easier to use yaml configuration
        if "apikey" in results["qsd"] and len(results["qsd"]["apikey"]):
            results["apikey"] = NotifyPagerDuty.unquote(
                results["qsd"]["apikey"]
            )
        else:
            results["apikey"] = NotifyPagerDuty.unquote(results["host"])

        # The 'integrationkey' makes it easier to use yaml configuration
        if "integrationkey" in results["qsd"] and len(
            results["qsd"]["integrationkey"]
        ):
            results["integrationkey"] = NotifyPagerDuty.unquote(
                results["qsd"]["integrationkey"]
            )
        else:
            results["integrationkey"] = NotifyPagerDuty.unquote(
                results["user"]
            )

        if "click" in results["qsd"] and len(results["qsd"]["click"]):
            results["click"] = NotifyPagerDuty.unquote(results["qsd"]["click"])

        if "group" in results["qsd"] and len(results["qsd"]["group"]):
            results["group"] = NotifyPagerDuty.unquote(results["qsd"]["group"])

        if "class" in results["qsd"] and len(results["qsd"]["class"]):
            results["class_id"] = NotifyPagerDuty.unquote(
                results["qsd"]["class"]
            )

        if "severity" in results["qsd"] and len(results["qsd"]["severity"]):
            results["severity"] = NotifyPagerDuty.unquote(
                results["qsd"]["severity"]
            )

        # Acquire our full path
        fullpath = NotifyPagerDuty.split_path(results["fullpath"])

        # Get our source
        if "source" in results["qsd"] and len(results["qsd"]["source"]):
            results["source"] = NotifyPagerDuty.unquote(
                results["qsd"]["source"]
            )
        else:
            results["source"] = fullpath.pop(0) if fullpath else None

        # Get our component
        if "component" in results["qsd"] and len(results["qsd"]["component"]):
            results["component"] = NotifyPagerDuty.unquote(
                results["qsd"]["component"]
            )
        else:
            results["component"] = fullpath.pop(0) if fullpath else None

        # Add our custom details key/value pairs that the user can potentially
        # over-ride if they wish to to our returned result set and tidy
        # entries by unquoting them
        results["details"] = {
            NotifyPagerDuty.unquote(x): NotifyPagerDuty.unquote(y)
            for x, y in results["qsd+"].items()
        }

        if "region" in results["qsd"] and len(results["qsd"]["region"]):
            # Extract from name to associate with from address
            results["region_name"] = NotifyPagerDuty.unquote(
                results["qsd"]["region"]
            )

        # Include images with our message
        results["include_image"] = parse_bool(
            results["qsd"].get("image", True)
        )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool
from .base import NotifyBase

# Default our global support flag
NOTIFY_GLIB_SUPPORT_ENABLED = False

# Image support is dependant on the GdkPixbuf library being available
NOTIFY_GLIB_IMAGE_SUPPORT = False


try:
    # glib essentials
    import gi
    gi.require_version("Gio", "2.0")
    gi.require_version("GLib", "2.0")
    from gi.repository import Gio, GLib

    # We're good
    NOTIFY_GLIB_SUPPORT_ENABLED = True

    # ImportError: When using gi.repository you must not import static modules
    # like "gobject". Please change all occurrences of "import gobject" to
    # "from gi.repository import GObject".
    # See: https://bugzilla.gnome.org/show_bug.cgi?id=709183
    if "gobject" in sys.modules:  # pragma: no cover
        del sys.modules["gobject"]

    try:
        # The following is required for Image/Icon loading only
        gi.require_version("GdkPixbuf", "2.0")
        from gi.repository import GdkPixbuf
        NOTIFY_GLIB_IMAGE_SUPPORT = True

    except (ImportError, ValueError, AttributeError):
        # No problem; this will get caught in outer try/catch

        # A ValueError will get thrown upon calling gi.require_version() if
        # GDK/GTK isn't installed on the system but gi is.
        pass

except ImportError:
    # No problem; we just simply can't support this plugin; we could
    # be in microsoft windows, or we just don't have the python-gobject
    # library available to us (or maybe one we don't support)?
    pass


# Urgencies
class GLibUrgency:
    LOW = 0
    NORMAL = 1
    HIGH = 2


GLIB_URGENCIES = {
    # Note: This also acts as a reverse lookup mapping
    GLibUrgency.LOW: "low",
    GLibUrgency.NORMAL: "normal",
    GLibUrgency.HIGH: "high",
}

GLIB_URGENCY_MAP = {
    # Maps against string 'low'
    "l": GLibUrgency.LOW,
    # Maps against string 'moderate'
    "m": GLibUrgency.LOW,
    # Maps against string 'normal'
    "n": GLibUrgency.NORMAL,
    # Maps against string 'high'
    "h": GLibUrgency.HIGH,
    # Maps against string 'emergency'
    "e": GLibUrgency.HIGH,

    # Entries to additionally support (so more like DBus's API)
    "0": GLibUrgency.LOW,
    "1": GLibUrgency.NORMAL,
    "2": GLibUrgency.HIGH,
}


class NotifyGLib(NotifyBase):
    """
    A wrapper for local GLib/Gio Notifications
    """

    # Set our global enabled flag
    enabled = NOTIFY_GLIB_SUPPORT_ENABLED

    requirements = {
        # Define our required packaging in order to work
        "details": _("libdbus-1.so.x or libdbus-2.so.x must be installed.")
    }

    # The default descriptive name associated with the Notification
    service_name = _("DBus Notification")

    # The services URL
    service_url = \
        "https://lazka.github.io/pgi-docs/Gio-2.0/classes/DBusProxy.html"

    # The default protocols
    protocol = ("glib", "gio")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_glib"

    # No throttling required for DBus queries
    request_rate_per_sec = 0

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_128

    # The number of milliseconds to keep the message present for
    message_timeout_ms = 13000

    # Limit results to just the first 10 line otherwise there is just to much
    # content to display
    body_max_line_count = 10

    # The following are required to hook into the notifications:
    glib_interface = "org.freedesktop.Notifications"
    glib_setting_location = "/org/freedesktop/Notifications"

    # Define object templates
    templates = (
        "{schema}://",
    )

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        "urgency": {
            "name": _("Urgency"),
            "type": "choice:int",
            "values": GLIB_URGENCIES,
            "default": GLibUrgency.NORMAL,
        },
        "priority": {
            # Apprise uses 'priority' everywhere; it's just a nice consistent
            # feel to be able to use it here as well. Just map the
            # value back to 'priority'
            "alias_of": "urgency",
        },
        "x": {
            "name": _("X-Axis"),
            "type": "int",
            "min": 0,
            "map_to": "x_axis",
        },
        "y": {
            "name": _("Y-Axis"),
            "type": "int",
            "min": 0,
            "map_to": "y_axis",
        },
        "image": {
            "name": _("Include Image"),
            "type": "bool",
            "default": True,
            "map_to": "include_image",
        },
    })

    def __init__(self, urgency=None, x_axis=None, y_axis=None,
                 include_image=True, **kwargs):
        """
        Initialize DBus Object
        """

        super().__init__(**kwargs)

        # Track our notifications
        self.registry = {}

        # The urgency of the message
        self.urgency = int(
            NotifyGLib.template_args["urgency"]["default"]
            if urgency is None else
            next((
                v for k, v in GLIB_URGENCY_MAP.items()
                if str(urgency).lower().startswith(k)),
                NotifyGLib.template_args["urgency"]["default"]))

        # Our x/y axis settings
        if x_axis or y_axis:
            try:
                self.x_axis = int(x_axis)
                self.y_axis = int(y_axis)

            except (TypeError, ValueError):
                # Invalid x/y values specified
                msg = "The x,y coordinates specified ({},{}) are invalid."\
                    .format(x_axis, y_axis)
                self.logger.warning(msg)
                raise TypeError(msg) from None
        else:
            self.x_axis = None
            self.y_axis = None

        # Track whether we want to add an image to the notification.
        self.include_image = include_image

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """
        Perform GLib/Gio Notification
        """
        # Acquire our gio interface
        try:
            gio_iface = Gio.DBusProxy.new_for_bus_sync(
                Gio.BusType.SESSION,
                Gio.DBusProxyFlags.NONE,
                None,
                self.glib_interface,
                self.glib_setting_location,
                self.glib_interface,
                None,
            )

        except GLib.Error as e:
            # Handle exception
            self.logger.warning("Failed to send GLib/Gio notification.")
            self.logger.debug(f"GLib/Gio Exception: {e}")
            return False

        # If there is no title, but there is a body, swap the two to get rid
        # of the weird whitespace
        if not title:
            title = body
            body = ""

        # image path
        icon_path = None if not self.include_image \
            else self.image_path(notify_type, extension=".ico")

        # Our meta payload
        meta_payload = {
            "urgency": GLib.Variant("y", self.urgency),
        }

        if not (self.x_axis is None and self.y_axis is None):
            # Set x/y access if these were set
            meta_payload["x"] = GLib.Variant("i", self.x_axis)
            meta_payload["y"] = GLib.Variant("i", self.y_axis)

        if NOTIFY_GLIB_IMAGE_SUPPORT and icon_path:
            try:
                # Use Pixbuf to create the proper image type
                image = GdkPixbuf.Pixbuf.new_from_file(icon_path)

                # Associate our image to our notification
                meta_payload["icon_data"] = GLib.Variant(
                    "(iiibiiay)",
                    (
                        image.get_width(),
                        image.get_height(),
                        image.get_rowstride(),
                        image.get_has_alpha(),
                        image.get_bits_per_sample(),
                        image.get_n_channels(),
                        image.get_pixels(),
                    ),
                )

            except Exception as e:
                self.logger.warning(
                    "Could not load notification icon (%s).", icon_path)
                self.logger.debug(f"GLib/Gio Exception: {e}")

        try:
            # Always call throttle() before any remote execution is made
            self.throttle()

            gio_iface.Notify(
                "(susssasa{sv}i)",
                # Application Identifier
                self.app_id,
                # Message ID (0 = New Message)
                0,
                # Icon (str) - not used
                "",
                # Title
                str(title),
                # Body
                str(body),
                # Actions
                [],
                # Meta
                meta_payload,
                # Message Timeout
                self.message_timeout_ms,
            )

            self.logger.info("Sent GLib/Gio notification.")

        except Exception as e:
            self.logger.warning("Failed to send GLib/Gio notification.")
            self.logger.debug(f"GLib/Gio Exception: {e}")
            return False

        return True

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
            "urgency":
                GLIB_URGENCIES[self.template_args["urgency"]["default"]]
                if self.urgency not in GLIB_URGENCIES
                else GLIB_URGENCIES[self.urgency],
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # x in (x,y) screen coordinates
        if self.x_axis:
            params["x"] = str(self.x_axis)

        # y in (x,y) screen coordinates
        if self.y_axis:
            params["y"] = str(self.y_axis)

        schema = self.protocol[0]
        return f"{schema}://_/?{NotifyGLib.urlencode(params)}"

    @staticmethod
    def parse_url(url):
        """
        There are no parameters necessary for this protocol; simply having
        gnome:// is all you need.  This function just makes sure that
        is in place.

        """

        results = NotifyBase.parse_url(url, verify_host=False)

        # Include images with our message
        results["include_image"] = \
            parse_bool(results["qsd"].get("image", True))

        # GLib/Gio supports urgency, but we we also support the keyword
        # priority so that it is consistent with some of the other plugins
        if "priority" in results["qsd"] and len(results["qsd"]["priority"]):
            # We intentionally store the priority in the urgency section
            results["urgency"] = \
                NotifyGLib.unquote(results["qsd"]["priority"])

        if "urgency" in results["qsd"] and len(results["qsd"]["urgency"]):
            results["urgency"] = \
                NotifyGLib.unquote(results["qsd"]["urgency"])

        # handle x,y coordinates
        if "x" in results["qsd"] and len(results["qsd"]["x"]):
            results["x_axis"] = NotifyGLib.unquote(results["qsd"].get("x"))

        if "y" in results["qsd"] and len(results["qsd"]["y"]):
            results["y_axis"] = NotifyGLib.unquote(results["qsd"].get("y"))

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# You must generate a "Long-Lived Access Token". This can be done from your
# Home Assistant Profile page.

from json import dumps
from uuid import uuid4

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import validate_regex
from .base import NotifyBase


class NotifyHomeAssistant(NotifyBase):
    """A wrapper for Home Assistant Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "HomeAssistant"

    # The services URL
    service_url = "https://www.home-assistant.io/"

    # Insecure Protocol Access
    protocol = "hassio"

    # Secure Protocol
    secure_protocol = "hassios"

    # Default to Home Assistant Default Insecure port of 8123 instead of 80
    default_insecure_port = 8123

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_homeassistant"

    # Define object templates
    templates = (
        "{schema}://{host}/{accesstoken}",
        "{schema}://{host}:{port}/{accesstoken}",
        "{schema}://{user}@{host}/{accesstoken}",
        "{schema}://{user}@{host}:{port}/{accesstoken}",
        "{schema}://{user}:{password}@{host}/{accesstoken}",
        "{schema}://{user}:{password}@{host}:{port}/{accesstoken}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "accesstoken": {
                "name": _("Long-Lived Access Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "nid": {
                # Optional Unique Notification ID
                "name": _("Notification ID"),
                "type": "string",
                "regex": (r"^[a-z0-9_-]+$", "i"),
            },
        },
    )

    def __init__(self, accesstoken, nid=None, **kwargs):
        """Initialize Home Assistant Object."""
        super().__init__(**kwargs)

        self.fullpath = kwargs.get("fullpath", "")

        if not (self.secure or self.port):
            # Use default insecure port
            self.port = self.default_insecure_port

        # Long-Lived Access token (generated from User Profile)
        self.accesstoken = validate_regex(accesstoken)
        if not self.accesstoken:
            msg = (
                "An invalid Home Assistant Long-Lived Access Token "
                f"({accesstoken}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # An Optional Notification Identifier
        self.nid = None
        if nid:
            self.nid = validate_regex(nid, *self.template_args["nid"]["regex"])
            if not self.nid:
                msg = (
                    "An invalid Home Assistant Notification Identifier "
                    f"({nid}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Sends Message."""

        # Prepare our persistent_notification.create payload
        payload = {
            "title": title,
            "message": body,
            # Use a unique ID so we don't over-write the last message
            # we posted. Otherwise use the notification id specified
            "notification_id": self.nid if self.nid else str(uuid4()),
        }

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.accesstoken}",
        }

        auth = None
        if self.user:
            auth = (self.user, self.password)

        # Set our schema
        schema = "https" if self.secure else "http"

        url = f"{schema}://{self.host}"
        if isinstance(self.port, int):
            url += f":{self.port}"

        url += (
            self.fullpath.rstrip("/")
            + "/api/services/persistent_notification/create"
        )

        self.logger.debug(
            "Home Assistant POST URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Home Assistant Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                url,
                data=dumps(payload),
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyHomeAssistant.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Home Assistant notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Home Assistant notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Home Assistant "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            (
                self.port
                if self.port
                else (443 if self.secure else self.default_insecure_port)
            ),
            self.fullpath.rstrip("/"),
            self.accesstoken,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {}
        if self.nid:
            params["nid"] = self.nid

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyHomeAssistant.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyHomeAssistant.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else self.default_insecure_port

        url = (
            "{schema}://{auth}{hostname}{port}{fullpath}"
            "{accesstoken}/?{params}"
        )

        return url.format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if not self.port or self.port == default_port
                else f":{self.port}"
            ),
            fullpath=(
                "/"
                if not self.fullpath
                else "/{}/".format(
                    NotifyHomeAssistant.quote(
                        self.fullpath.strip("/"), safe="/"
                    )
                )
            ),
            accesstoken=self.pprint(self.accesstoken, privacy, safe=""),
            params=NotifyHomeAssistant.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our Long-Lived Access Token
        if "accesstoken" in results["qsd"] and len(
            results["qsd"]["accesstoken"]
        ):
            results["accesstoken"] = NotifyHomeAssistant.unquote(
                results["qsd"]["accesstoken"]
            )

        else:
            # Acquire our full path
            fullpath = NotifyHomeAssistant.split_path(results["fullpath"])

            # Otherwise pop the last element from our path to be it
            results["accesstoken"] = fullpath.pop() if fullpath else None

            # Re-assemble our full path
            results["fullpath"] = "/" + "/".join(fullpath) if fullpath else ""

        # Allow the specification of a unique notification_id so that
        # it will always replace the last one sent.
        if "nid" in results["qsd"] and len(results["qsd"]["nid"]):
            results["nid"] = NotifyHomeAssistant.unquote(results["qsd"]["nid"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#
# API: https://dot.mindreset.tech/docs/service/studio/api/text_api
#      https://dot.mindreset.tech/docs/service/studio/api/image_api
#
# Text API Fields:
#   - refreshNow (bool, optional, default true): controls display timing.
#   - deviceId (string, required): unique device serial.
#   - title (string, optional): title text shown on screen.
#   - message (string, optional): body text shown on screen.
#   - signature (string, optional): footer/signature text.
#   - icon (string, optional): base64 PNG icon (40px x 40px).
#   - link (string, optional): tap-to-interact target URL.
#
# Image API Fields:
#   - refreshNow (bool, optional, default true): controls display timing.
#   - deviceId (string, required): unique device serial.
#   - image (string, required): base64 PNG image (296px x 152px).
#   - link (string, optional): tap-to-interact target URL.
#   - border (number, optional, default 0): 0=white, 1=black frame.
#   - ditherType (string, optional, default DIFFUSION): dithering mode.
#   - ditherKernel (string, optional, default FLOYD_STEINBERG):
#     dithering kernel.

from contextlib import suppress
import json

import requests

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool
from .base import NotifyBase

# Supported Dither Types
DOT_DITHER_TYPES = (
    "DIFFUSION",
    "ORDERED",
    "NONE",
)

# Supported Dither Kernels
DOT_DITHER_KERNELS = (
    "THRESHOLD",
    "ATKINSON",
    "BURKES",
    "FLOYD_STEINBERG",
    "SIERRA2",
    "STUCKI",
    "JARVIS_JUDICE_NINKE",
    "DIFFUSION_ROW",
    "DIFFUSION_COLUMN",
    "DIFFUSION_2D",
)


class NotifyDot(NotifyBase):
    """A wrapper for Dot. Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Dot."
    # Alias: devices marketed as "Quote/0" remain discoverable.

    # The services URL
    service_url = "https://dot.mindreset.tech"

    # All notification requests are secure
    secure_protocol = "dot"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_dot"

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_128

    # Support Attachments
    attachment_support = True

    # Supported API modes
    SUPPORTED_MODES = ("text", "image")

    DEFAULT_MODE = "text"

    # Define object templates
    templates = ("{schema}://{apikey}@{device_id}/{mode}/",)

    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "required": True,
                "private": True,
            },
            "device_id": {
                "name": _("Device Serial Number"),
                "type": "string",
                "required": True,
                "map_to": "device_id",
            },
            "mode": {
                "name": _("API Mode"),
                "type": "choice:string",
                "values": SUPPORTED_MODES,
                "default": DEFAULT_MODE,
                "map_to": "mode",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "refresh": {
                "name": _("Refresh Now"),
                "type": "bool",
                "default": True,
                "map_to": "refresh_now",
            },
            "signature": {
                "name": _("Text Signature"),
                "type": "string",
            },
            "icon": {
                "name": _("Icon Base64 (Text API)"),
                "type": "string",
            },
            "image": {
                "name": _("Image Base64 (Image API)"),
                "type": "string",
                "map_to": "image_data",
            },
            "link": {
                "name": _("Link"),
                "type": "string",
            },
            "border": {
                "name": _("Border"),
                "type": "int",
                "min": 0,
                "max": 1,
                "default": 0,
            },
            "dither_type": {
                "name": _("Dither Type"),
                "type": "choice:string",
                "values": DOT_DITHER_TYPES,
                "default": "DIFFUSION",
            },
            "dither_kernel": {
                "name": _("Dither Kernel"),
                "type": "choice:string",
                "values": DOT_DITHER_KERNELS,
                "default": "FLOYD_STEINBERG",
            },
        },
    )
    # Note:
    # - icon (Text API): base64 PNG icon (40px x 40px) in lower-left corner.
    #   Can be provided via `icon` parameter or first attachment.
    # - image (Image API): base64 PNG image (296px x 152px) supplied via
    #   configuration `image` parameter or first attachment.
    # - Only the first attachment is used; multiple attachments trigger a
    #   warning.

    def __init__(
        self,
        apikey=None,
        device_id=None,
        mode=DEFAULT_MODE,
        refresh_now=True,
        signature=None,
        icon=None,
        link=None,
        border=None,
        dither_type=None,
        dither_kernel=None,
        image_data=None,
        **kwargs,
    ):
        """Initialize Notify Dot Object."""
        super().__init__(**kwargs)

        # API Key (from user)
        self.apikey = apikey

        # Device ID tracks the Dot hardware serial.
        self.device_id = device_id

        # Refresh Now flag: True shows content immediately (default).
        self.refresh_now = parse_bool(refresh_now, default=True)

        # API mode ("text" or "image")
        self.mode = (
            mode.lower()
            if isinstance(mode, str) and mode.lower() in self.SUPPORTED_MODES
            else self.DEFAULT_MODE
        )
        if (
            not isinstance(mode, str)
            or mode.lower() not in self.SUPPORTED_MODES
        ):
            self.logger.warning(
                "Unsupported Dot mode (%s) specified; defaulting to '%s'.",
                mode,
                self.mode,
            )

        # Signature text used by the Text API footer.
        self.signature = signature if isinstance(signature, str) else None

        # Icon for the Text API (base64 PNG 40x40, lower-left corner).
        # Note: distinct from the Image API "image" field.
        self.icon = icon if isinstance(icon, str) else None

        # Image payload for the Image API (base64 PNG 296x152).
        self.image_data = image_data if isinstance(image_data, str) else None
        if self.mode == "text" and self.image_data:
            self.logger.warning(
                "Image data provided in text mode; ignoring configurable"
                " image payload."
            )
            self.image_data = None

        # Link for tap-to-interact navigation.
        self.link = link if isinstance(link, str) else None

        # Border for the Image API
        self.border = border

        # Dither type for Image API
        self.dither_type = dither_type

        # Dither kernel for the Image API
        self.dither_kernel = dither_kernel

        # Text API endpoint
        self.text_api_url = "https://dot.mindreset.tech/api/open/text"

        # Image API endpoint
        self.image_api_url = "https://dot.mindreset.tech/api/open/image"

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Dot Notification."""

        if not self.apikey:
            self.logger.warning("No API key was specified")
            return False

        if not self.device_id:
            self.logger.warning("No device ID was specified")
            return False

        # Prepare our headers
        headers = {
            "Authorization": f"Bearer {self.apikey}",
            "Content-Type": "application/json",
            "User-Agent": self.app_id,
        }

        if self.mode == "image":
            if title or body:
                self.logger.warning(
                    "Title and body are not supported in image mode "
                    "and will be ignored."
                )

            image_data = (
                self.image_data if isinstance(self.image_data, str) else None
            )

            # Use first attachment as image if no image_data provided
            # attachment.base64() returns base64-encoded string for API
            if not image_data and attach and self.attachment_support:
                if len(attach) > 1:
                    self.logger.warning(
                        "Multiple attachments provided; only the first "
                        "one will be used as image."
                    )
                try:
                    attachment = attach[0]
                    if attachment:
                        # Convert attachment to base64-encoded string
                        image_data = attachment.base64()
                except Exception as e:
                    self.logger.warning(f"Failed to process attachment: {e!s}")

            if not image_data:
                self.logger.warning(
                    "Image API mode selected but no image data was provided."
                )
                return False

            # Use Image API
            # Image API payload:
            #   refreshNow: display timing control.
            #   deviceId: Dot device serial (required).
            #   image: base64 PNG 296x152 (required).
            #   link: optional tap target.
            #   border: optional frame color.
            #   ditherType: optional dithering mode.
            #   ditherKernel: optional dithering kernel.
            payload = {
                "refreshNow": self.refresh_now,
                "deviceId": self.device_id,
                "image": image_data,  # Image payload shown on screen
            }

            if self.link:
                payload["link"] = self.link

            if self.border is not None:
                payload["border"] = self.border

            if self.dither_type is not None:
                payload["ditherType"] = self.dither_type

            if self.dither_kernel is not None:
                payload["ditherKernel"] = self.dither_kernel

            api_url = self.image_api_url

        else:
            # Use Text API
            # Text API payload:
            #   refreshNow: display timing control.
            #   deviceId: Dot device serial (required).
            #   title: optional title on screen.
            #   message: optional body on screen.
            #   signature: optional footer text.
            #   icon: optional base64 PNG icon (40x40).
            #   link: optional tap target.
            payload = {
                "refreshNow": self.refresh_now,
                "deviceId": self.device_id,
            }

            if title:
                payload["title"] = title

            if body:
                payload["message"] = body

            if self.signature:
                payload["signature"] = (
                    self.signature
                )  # Footer/signature displayed on screen

            # Use first attachment as icon if no icon provided
            # attachment.base64() returns base64-encoded string for API
            icon_data = self.icon
            if not icon_data and attach and self.attachment_support:
                if len(attach) > 1:
                    self.logger.warning(
                        "Multiple attachments provided; only the first "
                        "one will be used as icon."
                    )
                try:
                    attachment = attach[0]
                    if attachment:
                        # Convert attachment to base64-encoded string
                        icon_data = attachment.base64()
                except Exception as e:
                    self.logger.warning(f"Failed to process attachment: {e!s}")

            if icon_data:
                # Text API icon payload
                payload["icon"] = icon_data

            if self.link:
                payload["link"] = self.link

            api_url = self.text_api_url

        self.logger.debug(
            "Dot POST URL:"
            f" {api_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Dot Payload: {json.dumps(payload, indent=2)}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                api_url,
                data=json.dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code == requests.codes.ok:
                self.logger.info(f"Sent Dot notification to {self.device_id}.")
                return True

            # We had a problem
            status_str = NotifyDot.http_response_code_lookup(r.status_code)

            self.logger.warning(
                "Failed to send Dot notification to {}: "
                "{}{}error={}.".format(
                    self.device_id,
                    status_str,
                    ", " if status_str else "",
                    r.status_code,
                )
            )

            self.logger.debug(f"Response Details:\r\n{r.content}")

            return False

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Dot "
                f"notification to {self.device_id}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            return False

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another similar one.
        """
        return (
            self.secure_protocol,
            self.apikey,
            self.device_id,
            self.mode,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "refresh": "yes" if self.refresh_now else "no",
        }

        if self.mode == "text":
            if self.signature:
                params["signature"] = self.signature

            if self.icon:
                params["icon"] = self.icon

            if self.link:
                params["link"] = self.link

        else:  # image mode
            if self.image_data:
                params["image"] = self.image_data

            if self.link:
                params["link"] = self.link

            if self.border is not None:
                params["border"] = str(self.border)

            if self.dither_type and self.dither_type != "DIFFUSION":
                params["dither_type"] = self.dither_type

            if self.dither_kernel and self.dither_kernel != "FLOYD_STEINBERG":
                params["dither_kernel"] = self.dither_kernel

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        mode_segment = f"/{self.mode}/"

        return "{schema}://{apikey}@{device_id}{mode}?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(
                self.apikey, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            device_id=NotifyDot.quote(self.device_id, safe=""),
            mode=mode_segment,
            params=NotifyDot.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return 1 if self.device_id else 0

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Determine API mode from path (default to text)
        mode = NotifyDot.DEFAULT_MODE
        path_tokens = NotifyDot.split_path(results.get("fullpath"))
        if path_tokens:
            candidate = path_tokens.pop(0).lower()
            if candidate in NotifyDot.SUPPORTED_MODES:
                mode = candidate
            else:
                NotifyDot.logger.warning(
                    "Unsupported Dot mode (%s) detected; defaulting to '%s'.",
                    candidate,
                    NotifyDot.DEFAULT_MODE,
                )
        results["mode"] = mode
        remaining_path = "/".join(path_tokens)
        results["fullpath"] = "/" + remaining_path if remaining_path else "/"
        results["path"] = remaining_path

        # Extract API key from user
        user = results.get("user")
        if user:
            results["apikey"] = NotifyDot.unquote(user)

        # Extract device ID from hostname
        host = results.get("host")
        if host:
            results["device_id"] = NotifyDot.unquote(host)

        # Refresh Now
        refresh_value = results["qsd"].get("refresh")
        if refresh_value:
            results["refresh_now"] = parse_bool(refresh_value.strip())

        # Signature
        signature_value = results["qsd"].get("signature")
        if signature_value:
            results["signature"] = NotifyDot.unquote(signature_value.strip())

        # Icon
        icon_value = results["qsd"].get("icon")
        if icon_value:
            results["icon"] = NotifyDot.unquote(icon_value.strip())

        # Link
        link_value = results["qsd"].get("link")
        if link_value:
            results["link"] = NotifyDot.unquote(link_value.strip())

        # Border
        border_value = results["qsd"].get("border")
        if border_value:
            with suppress(TypeError, ValueError):
                results["border"] = int(border_value.strip())

        # Dither Type
        dither_type_value = results["qsd"].get("dither_type")
        if dither_type_value:
            results["dither_type"] = NotifyDot.unquote(
                dither_type_value.strip()
            )

        # Dither Kernel
        dither_kernel_value = results["qsd"].get("dither_kernel")
        if dither_kernel_value:
            results["dither_kernel"] = NotifyDot.unquote(
                dither_kernel_value.strip()
            )

        # Image (Image API)
        image_value = results["qsd"].get("image")
        if image_value:
            results["image_data"] = NotifyDot.unquote(image_value.strip())

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps, loads

import requests

from ..attachment.base import AttachBase
from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_email, parse_list, validate_regex
from .base import NotifyBase

# Flag used as a placeholder to sending to all devices
PUSHBULLET_SEND_TO_ALL = "ALL_DEVICES"

# Provide some known codes Pushbullet uses and what they translate to:
PUSHBULLET_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid Token.",
}


class NotifyPushBullet(NotifyBase):
    """A wrapper for PushBullet Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Pushbullet"

    # The services URL
    service_url = "https://www.pushbullet.com/"

    # The default secure protocol
    secure_protocol = "pbul"

    # Allow 50 requests per minute (Tier 2).
    # 60/50 = 0.2
    request_rate_per_sec = 1.2

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_pushbullet"

    # PushBullet uses the http protocol with JSON requests
    notify_url = "https://api.pushbullet.com/v2/{}"

    # Support attachments
    attachment_support = True

    # Define object templates
    templates = (
        "{schema}://{accesstoken}",
        "{schema}://{accesstoken}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "accesstoken": {
                "name": _("Access Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_device": {
                "name": _("Target Device"),
                "type": "string",
                "map_to": "targets",
            },
            "target_channel": {
                "name": _("Target Channel"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(self, accesstoken, targets=None, **kwargs):
        """Initialize PushBullet Object."""
        super().__init__(**kwargs)

        # Access Token (associated with project)
        self.accesstoken = validate_regex(accesstoken)
        if not self.accesstoken:
            msg = (
                "An invalid PushBullet Access Token "
                f"({accesstoken}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.targets = parse_list(targets)
        if len(self.targets) == 0:
            self.targets = (PUSHBULLET_SEND_TO_ALL,)

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform PushBullet Notification."""

        # error tracking (used for function return)
        has_error = False

        # Build a list of our attachments
        attachments = []

        if attach and self.attachment_support:
            # We need to upload our payload first so that we can source it
            # in remaining messages
            for no, attachment in enumerate(attach, start=1):

                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    "Preparing PushBullet attachment"
                    f" {attachment.url(privacy=True)}"
                )

                # prepare payload
                payload = {
                    "file_name": (
                        attachment.name
                        if attachment.name
                        else f"file{no:03}.dat"
                    ),
                    "file_type": attachment.mimetype,
                }
                # First thing we need to do is make a request so that we can
                # get a URL to post our request to.
                # see: https://docs.pushbullet.com/#upload-request
                okay, response = self._send(
                    self.notify_url.format("upload-request"), payload
                )
                if not okay:
                    # We can't post our attachment
                    return False

                # If we get here, our output will look something like this:
                # {
                #   "file_name": "cat.jpg",
                #   "file_type": "image/jpeg",
                #   "file_url": "https://dl.pushb.com/abc/cat.jpg",
                #   "upload_url": "https://upload.pushbullet.com/abcd123"
                # }

                # - The file_url is where the file will be available after it
                #    is uploaded.
                # - The upload_url is where to POST the file to. The file must
                #    be posted using multipart/form-data encoding.

                # Prepare our attachment payload; we'll use this if we
                # successfully upload the content below for later on.
                try:
                    # By placing this in a try/except block we can validate
                    # our response at the same time as preparing our payload
                    payload = {
                        # PushBullet v2/pushes file type:
                        "type": "file",
                        "file_name": response["file_name"],
                        "file_type": response["file_type"],
                        "file_url": response["file_url"],
                    }

                    if response["file_type"].startswith("image/"):
                        # Allow image to be displayed inline (if image type)
                        payload["image_url"] = response["file_url"]

                    upload_url = response["upload_url"]

                except (KeyError, TypeError):
                    # A method of verifying our content exists
                    return False

                okay, response = self._send(upload_url, attachment)
                if not okay:
                    # We can't post our attachment
                    return False

                # Save our pre-prepared payload for attachment posting
                attachments.append(payload)

        # Create a copy of the targets list
        targets = list(self.targets)
        while len(targets):
            recipient = targets.pop(0)

            # prepare payload
            payload = {
                "type": "note",
                "title": title,
                "body": body,
            }

            # Check if an email was defined
            match = is_email(recipient)
            if match:
                payload["email"] = match["full_email"]
                self.logger.debug(
                    f"PushBullet recipient {recipient} parsed as an email"
                    " address"
                )

            elif recipient is PUSHBULLET_SEND_TO_ALL:
                # Send to all
                pass

            elif recipient[0] == "#":
                payload["channel_tag"] = recipient[1:]
                self.logger.debug(
                    f"PushBullet recipient {recipient} parsed as a channel"
                )

            else:
                payload["device_iden"] = recipient
                self.logger.debug(
                    f"PushBullet recipient {recipient} parsed as a device"
                )

            if body:
                okay, response = self._send(
                    self.notify_url.format("pushes"), payload
                )
                if not okay:
                    has_error = True
                    continue

                self.logger.info(
                    f'Sent PushBullet notification to "{recipient}".'
                )

            for attach_payload in attachments:
                # Send our attachments to our same user (already prepared as
                # our payload object)
                okay, response = self._send(
                    self.notify_url.format("pushes"), attach_payload
                )
                if not okay:
                    has_error = True
                    continue

                self.logger.info(
                    'Sent PushBullet attachment ({}) to "{}".'.format(
                        attach_payload["file_name"], recipient
                    )
                )

        return not has_error

    def _send(self, url, payload, **kwargs):
        """Wrapper to the requests (post) object."""

        headers = {
            "User-Agent": self.app_id,
        }

        # Some default values for our request object to which we'll update
        # depending on what our payload is
        files = None
        data = None

        if not isinstance(payload, AttachBase):
            # Send our payload as a JSON object
            headers["Content-Type"] = "application/json"
            data = dumps(payload) if payload else None

        auth = (self.accesstoken, "")

        self.logger.debug(
            "PushBullet POST URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"PushBullet Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        # Default response type
        response = None

        try:
            # Open our attachment path if required:
            if isinstance(payload, AttachBase):
                files = {
                    "file": (
                        payload.name,
                        # file handle is safely closed in `finally`; inline
                        # open is intentional
                        open(payload.path, "rb"),  # noqa: SIM115
                    )}

            r = requests.post(
                url,
                data=data,
                headers=headers,
                files=files,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            try:
                response = loads(r.content)

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None

                # Fall back to the existing unparsed value
                response = r.content

            if r.status_code not in (
                requests.codes.ok,
                requests.codes.no_content,
            ):
                # We had a problem
                status_str = NotifyPushBullet.http_response_code_lookup(
                    r.status_code, PUSHBULLET_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to deliver payload to PushBullet:"
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                return False, response

            # otherwise we were successful
            return True, response

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred communicating with PushBullet."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            return False, response

        except OSError as e:
            self.logger.warning(
                "An I/O error occurred while handling {}.".format(
                    payload.name
                    if isinstance(payload, AttachBase)
                    else payload
                )
            )
            self.logger.debug(f"I/O Exception: {e!s}")
            return False, response

        finally:
            # Close our file (if it's open) stored in the second element
            # of our files tuple (index 1)
            if files:
                files["file"][1].close()

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.accesstoken)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        targets = "/".join([NotifyPushBullet.quote(x) for x in self.targets])
        if targets == PUSHBULLET_SEND_TO_ALL:
            # keyword is reserved for internal usage only; it's safe to remove
            # it from the recipients list
            targets = ""

        return "{schema}://{accesstoken}/{targets}/?{params}".format(
            schema=self.secure_protocol,
            accesstoken=self.pprint(self.accesstoken, privacy, safe=""),
            targets=targets,
            params=NotifyPushBullet.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Fetch our targets
        results["targets"] = NotifyPushBullet.split_path(results["fullpath"])

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyPushBullet.parse_list(
                results["qsd"]["to"]
            )

        # Setup the token; we store it in Access Token for global
        # plugin consistency with naming conventions
        results["accesstoken"] = NotifyPushBullet.unquote(results["host"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Notifico allows you to relay notifications into IRC channels.
#
# 1. visit https://n.tkte.ch and sign up for an account
# 2. create a project; either manually or sync with github
# 3. from within the project, you can create a message hook
#
# the URL will look something like this:
#       https://n.tkte.ch/h/2144/uJmKaBW9WFk42miB146ci3Kj
#                            ^                ^
#                            |                |
#                         project id       message hook
#
# This plugin also supports taking the URL (as identified above) directly
# as well.

import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool, validate_regex
from .base import NotifyBase


class NotificoFormat:
    # Resets all formatting
    Reset = "\x0f"

    # Formatting
    Bold = "\x02"
    Italic = "\x1d"
    Underline = "\x1f"
    BGSwap = "\x16"


class NotificoColor:
    # Resets Color
    Reset = "\x03"

    # Colors
    White = "\x0300"
    Black = "\x0301"
    Blue = "\x0302"
    Green = "\x0303"
    Red = "\x0304"
    Brown = "\x0305"
    Purple = "\x0306"
    Orange = "\x0307"
    Yellow = ("\x0308",)
    LightGreen = "\x0309"
    Teal = "\x0310"
    LightCyan = "\x0311"
    LightBlue = "\x0312"
    Violet = "\x0313"
    Grey = "\x0314"
    LightGrey = "\x0315"


class NotifyNotifico(NotifyBase):
    """A wrapper for Notifico Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Notifico"

    # The services URL
    service_url = "https://n.tkte.ch"

    # The default protocol
    protocol = "notifico"

    # The default secure protocol
    secure_protocol = "notifico"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_notifico"

    # Plain Text Notification URL
    notify_url = "https://n.tkte.ch/h/{proj}/{hook}"

    # The title is not used
    title_maxlen = 0

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 512

    # Define object templates
    templates = ("{schema}://{project_id}/{msghook}",)

    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            # The Project ID is found as the first part of the URL
            #  /1234/........................
            "project_id": {
                "name": _("Project ID"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[0-9]+$", ""),
            },
            # The Message Hook follows the Project ID
            #  /..../AbCdEfGhIjKlMnOpQrStUvWX
            "msghook": {
                "name": _("Message Hook"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[a-z0-9]+$", "i"),
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            # You can optionally pass IRC colors into
            "color": {
                "name": _("IRC Colors"),
                "type": "bool",
                "default": True,
            },
            # You can optionally pass IRC color into
            "prefix": {
                "name": _("Prefix"),
                "type": "bool",
                "default": True,
            },
        },
    )

    def __init__(self, project_id, msghook, color=True, prefix=True, **kwargs):
        """Initialize Notifico Object."""
        super().__init__(**kwargs)

        # Assign our message hook
        self.project_id = validate_regex(
            project_id, *self.template_tokens["project_id"]["regex"]
        )
        if not self.project_id:
            msg = (
                f"An invalid Notifico Project ID ({project_id}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Assign our message hook
        self.msghook = validate_regex(
            msghook, *self.template_tokens["msghook"]["regex"]
        )
        if not self.msghook:
            msg = (
                f"An invalid Notifico Message Token ({msghook}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Prefix messages with a [?] where ? identifies the message type
        # such as if it's an error, warning, info, or success
        self.prefix = prefix

        # Send colors
        self.color = color

        # Prepare our notification URL now:
        self.api_url = self.notify_url.format(
            proj=self.project_id,
            hook=self.msghook,
        )
        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.project_id, self.msghook)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "color": "yes" if self.color else "no",
            "prefix": "yes" if self.prefix else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{proj}/{hook}/?{params}".format(
            schema=self.secure_protocol,
            proj=self.pprint(self.project_id, privacy, safe=""),
            hook=self.pprint(self.msghook, privacy, safe=""),
            params=NotifyNotifico.urlencode(params),
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Wrapper to _send since we can alert more then one channel."""

        # prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded; charset=utf-8",
        }

        # Prepare our IRC Prefix
        color = ""
        token = ""
        if notify_type == NotifyType.INFO:
            color = NotificoColor.Teal
            token = "i"

        elif notify_type == NotifyType.SUCCESS:
            color = NotificoColor.LightGreen
            token = "✔"

        elif notify_type == NotifyType.WARNING:
            color = NotificoColor.Orange
            token = "!"

        elif notify_type == NotifyType.FAILURE:
            color = NotificoColor.Red
            token = "✗"

        if self.color:
            # Colors were specified, make sure we capture and correctly
            # allow them to exist inline in the message
            # \g<1> is less ambiguous than \1
            body = re.sub(r"\\x03(\d{0,2})", r"\\x03\g<1>", body)

        else:
            # no colors specified, make sure we strip out any colors found
            # to make the string read-able
            body = re.sub(r"\\x03(\d{1,2}(,[0-9]{1,2})?)?", r"", body)

        # Prepare our payload
        payload = {
            "payload": (
                body
                if not self.prefix
                else "{}[{}]{} {}{}{}: {}{}".format(
                    # Token [?] at the head
                    color if self.color else "",
                    token,
                    NotificoColor.Reset if self.color else "",
                    # App ID
                    NotificoFormat.Bold if self.color else "",
                    self.app_id,
                    NotificoFormat.Reset if self.color else "",
                    # Message Body
                    body,
                    # Reset
                    NotificoFormat.Reset if self.color else "",
                )
            ),
        }

        self.logger.debug(
            "Notifico GET URL:"
            f" {self.api_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Notifico Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.get(
                self.api_url,
                params=payload,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyNotifico.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Notifico notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Notifico notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Notifico notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The first token is stored in the hostname
        results["project_id"] = NotifyNotifico.unquote(results["host"])

        # Get Message Hook
        try:
            results["msghook"] = NotifyNotifico.split_path(
                results["fullpath"]
            )[0]

        except IndexError:
            results["msghook"] = None

        # Include Color
        results["color"] = parse_bool(results["qsd"].get("color", True))

        # Include Prefix
        results["prefix"] = parse_bool(results["qsd"].get("prefix", True))

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://n.tkte.ch/h/PROJ_ID/MESSAGE_HOOK/
        """

        result = re.match(
            r"^https?://n\.tkte\.ch/h/"
            r"(?P<proj>[0-9]+)/"
            r"(?P<hook>[A-Z0-9]+)/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            return NotifyNotifico.parse_url(
                "{schema}://{proj}/{hook}/{params}".format(
                    schema=NotifyNotifico.secure_protocol,
                    proj=result.group("proj"),
                    hook=result.group("hook"),
                    params=(
                        ""
                        if not result.group("params")
                        else result.group("params")
                    ),
                )
            )

        return None



#
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Assumes QQ Push API provided by third-party bridge like message-pusher

import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import validate_regex
from .base import NotifyBase


class NotifyQQ(NotifyBase):
    """A wrapper for QQ Push Notifications."""

    # The default descriptive name associated with the Notification
    service_name = _("QQ Push")

    # The services URL
    service_url = "https://github.com/songquanpeng/message-pusher"

    # The default secure protocol
    secure_protocol = "qq"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_qq"

    # URL used to send notifications with
    notify_url = "https://qmsg.zendee.cn/send/"

    templates = ("{schema}://{token}",)

    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("User Token"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9]{24,64}$", "i"),
            },
        },
    )

    def __init__(self, token, **kwargs):
        """Initialize QQ Push Object.

        Args:
            token (str): User push token from QQ Push provider (e.g., Qmsg)
        """
        super().__init__(**kwargs)

        self.token = validate_regex(
            token, *self.template_tokens["token"]["regex"]
        )
        if not self.token:
            msg = f"The QQ Push token ({token}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.webhook_url = f"{self.notify_url}{self.token}"

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""
        params = self.url_parameters(privacy=privacy, *args, **kwargs)
        return (
            f"{self.secure_protocol}://"
            f"{self.pprint(self.token, privacy, mode=PrivacyMode.Secret)}/"
            f"?{self.urlencode(params)}"
        )

    @property
    def url_identifier(self):
        """Returns a unique identifier for this plugin instance."""
        return (self.secure_protocol, self.token)

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Send a QQ Push Notification."""
        payload = {"msg": f"{title}\n{body}" if title else body}

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded",
        }

        self.throttle()
        try:
            response = requests.post(
                self.webhook_url,
                headers=headers,
                data=payload,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if response.status_code != requests.codes.ok:
                self.logger.warning(
                    "QQ Push notification failed: %d - %s",
                    response.status_code,
                    response.text,
                )
                return False

        except requests.RequestException as e:
            self.logger.warning(f"QQ Push Exception: {e}")
            return False

        self.logger.info("QQ Push notification sent successfully.")
        return True

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns arguments to re-instantiate the
        object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            return results

        if "token" in results["qsd"] and results["qsd"]["token"]:
            results["token"] = NotifyQQ.unquote(results["qsd"]["token"])
        else:
            results["token"] = NotifyQQ.unquote(results["host"])

        return results

    @staticmethod
    def parse_native_url(url):
        """Parse native QQ push-style URL into Apprise format."""
        match = re.match(
            r"^https://qmsg\.zendee\.cn/send/([a-z0-9]+)$", url, re.I
        )
        if not match:
            return None

        return NotifyQQ.parse_url(
            f"{NotifyQQ.secure_protocol}://{match.group(1)}"
        )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps
from uuid import uuid4

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase


# Actions
class PagerTreeAction:
    CREATE = "create"
    ACKNOWLEDGE = "acknowledge"
    RESOLVE = "resolve"


# Urgencies
class PagerTreeUrgency:
    SILENT = "silent"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    CRITICAL = "critical"


PAGERTREE_ACTIONS = {
    PagerTreeAction.CREATE: "create",
    PagerTreeAction.ACKNOWLEDGE: "acknowledge",
    PagerTreeAction.RESOLVE: "resolve",
}

PAGERTREE_URGENCIES = {
    # Note: This also acts as a reverse lookup mapping
    PagerTreeUrgency.SILENT: "silent",
    PagerTreeUrgency.LOW: "low",
    PagerTreeUrgency.MEDIUM: "medium",
    PagerTreeUrgency.HIGH: "high",
    PagerTreeUrgency.CRITICAL: "critical",
}
# Extend HTTP Error Messages
PAGERTREE_HTTP_ERROR_MAP = {
    402: "Payment Required - Please subscribe or upgrade",
    403: "Forbidden - Blocked",
    404: "Not Found - Invalid Integration ID",
    405: "Method Not Allowed - Integration Disabled",
    429: "Too Many Requests - Rate Limit Exceeded",
}


class NotifyPagerTree(NotifyBase):
    """A wrapper for PagerTree Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "PagerTree"

    # The services URL
    service_url = "https://pagertree.com/"

    # All PagerTree requests are secure
    secure_protocol = "pagertree"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_pagertree"

    # PagerTree uses the http protocol with JSON requests
    notify_url = "https://api.pagertree.com/integration/{}"

    # Define object templates
    templates = ("{schema}://{integration}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "integration": {
                "name": _("Integration ID"),
                "type": "string",
                "private": True,
                "required": True,
            }
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "action": {
                "name": _("Action"),
                "type": "choice:string",
                "values": PAGERTREE_ACTIONS,
                "default": PagerTreeAction.CREATE,
            },
            "thirdparty": {
                "name": _("Third Party ID"),
                "type": "string",
            },
            "urgency": {
                "name": _("Urgency"),
                "type": "choice:string",
                "values": PAGERTREE_URGENCIES,
            },
            "tags": {
                "name": _("Tags"),
                "type": "string",
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "headers": {
            "name": _("HTTP Header"),
            "prefix": "+",
        },
        "payload_extras": {
            "name": _("Payload Extras"),
            "prefix": ":",
        },
        "meta_extras": {
            "name": _("Meta Extras"),
            "prefix": "-",
        },
    }

    def __init__(
        self,
        integration,
        action=None,
        thirdparty=None,
        urgency=None,
        tags=None,
        headers=None,
        payload_extras=None,
        meta_extras=None,
        **kwargs,
    ):
        """Initialize PagerTree Object."""
        super().__init__(**kwargs)

        # Integration ID (associated with account)
        self.integration = validate_regex(
            integration, r"^int_[a-zA-Z0-9\-_]{7,14}$"
        )
        if not self.integration:
            msg = (
                "An invalid PagerTree Integration ID "
                f"({integration}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # thirdparty (optional, in case they want to pass the
        # acknowledge or resolve action)
        self.thirdparty = None
        if thirdparty:
            # An id was specified, we want to validate it
            self.thirdparty = validate_regex(thirdparty)
            if not self.thirdparty:
                msg = (
                    "An invalid PagerTree third party ID "
                    f"({thirdparty}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        self.payload_extras = {}
        if payload_extras:
            # Store our extra payload entries
            self.payload_extras.update(payload_extras)

        self.meta_extras = {}
        if meta_extras:
            # Store our extra payload entries
            self.meta_extras.update(meta_extras)

        # Setup our action
        self.action = (
            NotifyPagerTree.template_args["action"]["default"]
            if action not in PAGERTREE_ACTIONS
            else PAGERTREE_ACTIONS[action]
        )

        # Setup our urgency
        self.urgency = PAGERTREE_URGENCIES.get(urgency)

        # Any optional tags to attach to the notification
        self.__tags = parse_list(tags)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform PagerTree Notification."""

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Apply any/all header over-rides defined
        # For things like PagerTree Token
        headers.update(self.headers)

        # prepare JSON Object
        payload = {
            # Generate an ID (unless one was explicitly forced to be used)
            "id": self.thirdparty if self.thirdparty else str(uuid4()),
            "event_type": self.action,
        }

        if self.action == PagerTreeAction.CREATE:
            payload["title"] = title if title else self.app_desc
            payload["description"] = body

            payload["meta"] = self.meta_extras
            payload["tags"] = self.__tags

            if self.urgency is not None:
                payload["urgency"] = self.urgency

        # Apply any/all payload over-rides defined
        payload.update(self.payload_extras)

        # Prepare our URL based on integration
        notify_url = self.notify_url.format(self.integration)

        self.logger.debug(
            "PagerTree POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"PagerTree Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                notify_url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code not in (
                requests.codes.ok,
                requests.codes.created,
                requests.codes.accepted,
            ):
                # We had a problem
                status_str = NotifyPagerTree.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send PagerTree notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent PagerTree notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending PagerTree "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.integration)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "action": self.action,
        }

        if self.thirdparty:
            params["tid"] = self.thirdparty

        if self.urgency:
            params["urgency"] = self.urgency

        if self.__tags:
            params["tags"] = ",".join(list(self.__tags))

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Headers prefixed with a '+' sign
        # Append our headers into our parameters
        params.update({f"+{k}": v for k, v in self.headers.items()})

        # Meta: {} prefixed with a '-' sign
        # Append our meta extras into our parameters
        params.update({f"-{k}": v for k, v in self.meta_extras.items()})

        # Payload body extras prefixed with a ':' sign
        # Append our payload extras into our parameters
        params.update({f":{k}": v for k, v in self.payload_extras.items()})

        return "{schema}://{integration}?{params}".format(
            schema=self.secure_protocol,
            # never encode hostname since we're expecting it to be a valid one
            integration=self.pprint(self.integration, privacy, safe=""),
            params=NotifyPagerTree.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)

        if not results:
            # We're done early as we couldn't load the results
            return results

        # Add our headers that the user can potentially over-ride if they wish
        # to to our returned result set and tidy entries by unquoting them
        results["headers"] = {
            NotifyPagerTree.unquote(x): NotifyPagerTree.unquote(y)
            for x, y in results["qsd+"].items()
        }

        # store any additional payload extra's defined
        results["payload_extras"] = {
            NotifyPagerTree.unquote(x): NotifyPagerTree.unquote(y)
            for x, y in results["qsd:"].items()
        }

        # store any additional meta extra's defined
        results["meta_extras"] = {
            NotifyPagerTree.unquote(x): NotifyPagerTree.unquote(y)
            for x, y in results["qsd-"].items()
        }

        # Integration ID
        if "id" in results["qsd"] and len(results["qsd"]["id"]):
            # Shortened version of integration id
            results["integration"] = NotifyPagerTree.unquote(
                results["qsd"]["id"]
            )

        elif "integration" in results["qsd"] and len(
            results["qsd"]["integration"]
        ):
            results["integration"] = NotifyPagerTree.unquote(
                results["qsd"]["integration"]
            )

        else:
            results["integration"] = NotifyPagerTree.unquote(results["host"])

        # Set our thirdparty

        if "tid" in results["qsd"] and len(results["qsd"]["tid"]):
            # Shortened version of thirdparty
            results["thirdparty"] = NotifyPagerTree.unquote(
                results["qsd"]["tid"]
            )

        elif "thirdparty" in results["qsd"] and len(
            results["qsd"]["thirdparty"]
        ):
            results["thirdparty"] = NotifyPagerTree.unquote(
                results["qsd"]["thirdparty"]
            )

        # Set our urgency
        if "action" in results["qsd"] and len(results["qsd"]["action"]):
            results["action"] = NotifyPagerTree.unquote(
                results["qsd"]["action"]
            )

        # Set our urgency
        if "urgency" in results["qsd"] and len(results["qsd"]["urgency"]):
            results["urgency"] = NotifyPagerTree.unquote(
                results["qsd"]["urgency"]
            )

        # Set our tags
        if "tags" in results["qsd"] and len(results["qsd"]["tags"]):
            results["tags"] = parse_list(
                NotifyPagerTree.unquote(results["qsd"]["tags"])
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Create an incoming webhook; the website will provide you with something like:
#  http://localhost:8065/hooks/yobjmukpaw3r3urc5h6i369yima
#                              ^^^^^^^^^^^^^^^^^^^^^^^^^^^
#                              |-- this is the webhook --|
#
# You can effectively turn the url above to read this:
# mmost://localhost:8065/yobjmukpaw3r3urc5h6i369yima
#  - swap http with mmost
#  - drop /hooks/ reference

from json import dumps
import re

import requests

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_bool, parse_list, validate_regex
from .base import NotifyBase

# Some Reference Locations:
# - https://docs.mattermost.com/developer/webhooks-incoming.html
# - https://docs.mattermost.com/administration/config-settings.html


class NotifyMattermost(NotifyBase):
    """A wrapper for Mattermost Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Mattermost"

    # The services URL
    service_url = "https://mattermost.com/"

    # The default protocol
    protocol = "mmost"

    # The default secure protocol
    secure_protocol = "mmosts"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_mattermost"

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_72

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 4000

    # Mattermost does not have a title
    title_maxlen = 0

    # Define object templates
    templates = (
        "{schema}://{host}/{token}",
        "{schema}://{host}:{port}/{token}",
        "{schema}://{host}/{fullpath}/{token}",
        "{schema}://{host}:{port}/{fullpath}/{token}",
        "{schema}://{botname}@{host}/{token}",
        "{schema}://{botname}@{host}:{port}/{token}",
        "{schema}://{botname}@{host}/{fullpath}/{token}",
        "{schema}://{botname}@{host}:{port}/{fullpath}/{token}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "token": {
                "name": _("Webhook Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "fullpath": {
                "name": _("Path"),
                "type": "string",
            },
            "botname": {
                "name": _("Bot Name"),
                "type": "string",
                "map_to": "user",
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "channels": {
                "name": _("Channels"),
                "type": "list:string",
            },
            "channel": {
                "alias_of": "channels",
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": True,
                "map_to": "include_image",
            },
            "to": {
                "alias_of": "channels",
            },
        },
    )

    def __init__(
        self,
        token,
        fullpath=None,
        channels=None,
        include_image=False,
        **kwargs,
    ):
        """Initialize Mattermost Object."""
        super().__init__(**kwargs)

        if self.secure:
            self.schema = "https"

        else:
            self.schema = "http"

        # our full path
        self.fullpath = (
            "" if not isinstance(fullpath, str) else fullpath.strip()
        )

        # Authorization Token (associated with project)
        self.token = validate_regex(token)
        if not self.token:
            msg = (
                "An invalid Mattermost Authorization Token "
                f"({token}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Optional Channels (strip off any channel prefix entries if present)
        self.channels = [x.lstrip("#") for x in parse_list(channels)]

        # Place a thumbnail image inline with the message body
        self.include_image = include_image

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Mattermost Notification."""

        # Create a copy of our channels, otherwise place a dummy entry
        channels = (
            list(self.channels)
            if self.channels
            else [
                None,
            ]
        )

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # prepare JSON Object
        payload = {
            "text": body,
            "icon_url": None,
        }

        # Acquire our image url if configured to do so
        image_url = (
            None if not self.include_image else self.image_url(notify_type)
        )

        if image_url:
            # Set our image configuration if told to do so
            payload["icon_url"] = image_url

        # Set our user
        payload["username"] = self.user if self.user else self.app_id

        port = ""
        if self.port is not None:
            port = f":{self.port}"

        # For error tracking
        has_error = False

        while len(channels):
            # Pop a channel off of the list
            channel = channels.pop(0)

            if channel:
                payload["channel"] = channel

            url = "{}://{}{}{}/hooks/{}".format(
                self.schema,
                self.host,
                port,
                self.fullpath.rstrip("/"),
                self.token,
            )

            self.logger.debug(
                "Mattermost POST URL:"
                f" {url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"Mattermost Payload: {payload!s}")

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.post(
                    url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyMattermost.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Mattermost notification{}: "
                        "{}{}error={}.".format(
                            "" if not channel else f" to channel {channel}",
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Flag our error
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        "Sent Mattermost notification{}.".format(
                            "" if not channel else f" to channel {channel}"
                        )
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Mattermost "
                    "notification{}.".format(
                        "" if not channel else f" to channel {channel}"
                    )
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Flag our error
                has_error = True
                continue

        # Return our overall status
        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.token,
            self.host,
            self.port,
            self.fullpath,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        if self.channels:
            # historically the value only accepted one channel and is
            # therefore identified as 'channel'. Channels have always been
            # optional, so that is why this setting is nested in an if block
            params["channel"] = ",".join(
                [NotifyMattermost.quote(x, safe="") for x in self.channels]
            )

        default_port = 443 if self.secure else 80
        default_schema = self.secure_protocol if self.secure else self.protocol

        # Determine if there is a botname present
        botname = ""
        if self.user:
            botname = "{botname}@".format(
                botname=NotifyMattermost.quote(self.user, safe=""),
            )

        return (
            "{schema}://{botname}{hostname}{port}{fullpath}{token}"
            "/?{params}".format(
                schema=default_schema,
                botname=botname,
                # never encode hostname since we're expecting it to be a valid
                # one
                hostname=self.host,
                port=(
                    ""
                    if self.port is None or self.port == default_port
                    else f":{self.port}"
                ),
                fullpath=(
                    "/"
                    if not self.fullpath
                    else "{}/".format(
                        NotifyMattermost.quote(self.fullpath, safe="/")
                    )
                ),
                token=self.pprint(self.token, privacy, safe=""),
                params=NotifyMattermost.urlencode(params),
            )
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Acquire our tokens; the last one will always be our token
        # all entries before it will be our path
        tokens = NotifyMattermost.split_path(results["fullpath"])

        results["token"] = None if not tokens else tokens.pop()

        # Store our path
        results["fullpath"] = (
            "" if not tokens else "/{}".format("/".join(tokens))
        )

        # Define our optional list of channels to notify
        results["channels"] = []

        # Support both 'to' (for yaml configuration) and channel=
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            # Allow the user to specify the channel to post to
            results["channels"].extend(
                NotifyMattermost.parse_list(results["qsd"]["to"])
            )

        if "channel" in results["qsd"] and len(results["qsd"]["channel"]):
            # Allow the user to specify the channel to post to
            results["channels"].extend(
                NotifyMattermost.parse_list(results["qsd"]["channel"])
            )

        if "channels" in results["qsd"] and len(results["qsd"]["channels"]):
            # Allow the user to specify the channel to post to
            results["channels"].extend(
                NotifyMattermost.parse_list(results["qsd"]["channels"])
            )

        # Image manipulation
        results["include_image"] = parse_bool(
            results["qsd"].get(
                "image", NotifyMattermost.template_args["image"]["default"]
            )
        )

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support parsing the webhook straight from URL
            https://HOST:443/workflows/WORKFLOWID/triggers/manual/paths/invoke
            https://mattermost.HOST/hooks/TOKEN
        """

        # Match our workflows webhook URL and re-assemble
        result = re.match(
            r"^http(?P<secure>s?)://(?P<host>mattermost\.[A-Z0-9_.-]+)"
            r"(:(?P<port>[1-9][0-9]{0,5}))?"
            r"/hooks/"
            r"(?P<token>[A-Z0-9_-]+)/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            default_port = (
                int(result.group("port"))
                if result.group("port")
                else (443 if result.group("secure") else 80)
            )

            default_schema = (
                NotifyMattermost.secure_protocol
                if result.group("secure")
                else NotifyMattermost.protocol
            )

            # Construct our URL
            return NotifyMattermost.parse_url(
                "{schema}://{host}{port}/{token}/{params}".format(
                    schema=default_schema,
                    host=result.group("host"),
                    port=(
                        ""
                        if not result.group("port")
                        or int(result.group("port")) == default_port
                        else f":{default_port}"
                    ),
                    token=result.group("token"),
                    params=(
                        ""
                        if not result.group("params")
                        else result.group("params")
                    ),
                )
            )
        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# One Signal requires that you've signed up with the service and
# generated yourself an API Key and APP ID.

# Sources:
#  - https://documentation.onesignal.com/docs/accounts-and-keys
#  - https://documentation.onesignal.com/reference/create-notification

from itertools import chain
from json import dumps

import requests

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..utils.base64 import decode_b64_dict, encode_b64_dict
from ..utils.parse import is_email, parse_bool, parse_list, validate_regex
from .base import NotifyBase


class OneSignalCategory:
    """We define the different category types that we can notify via
    OneSignal."""

    PLAYER = "include_player_ids"
    EMAIL = "include_email_tokens"
    USER = "include_external_user_ids"
    SEGMENT = "included_segments"


ONESIGNAL_CATEGORIES = (
    OneSignalCategory.PLAYER,
    OneSignalCategory.EMAIL,
    OneSignalCategory.USER,
    OneSignalCategory.SEGMENT,
)


class NotifyOneSignal(NotifyBase):
    """A wrapper for OneSignal Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "OneSignal"

    # The services URL
    service_url = "https://onesignal.com"

    # The default protocol
    secure_protocol = "onesignal"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_onesignal"

    # Notification
    notify_url = "https://api.onesignal.com/notifications"

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_72

    # The maximum allowable batch sizes per message
    default_batch_size = 2000

    # Define object templates
    templates = (
        "{schema}://{app}@{apikey}/{targets}",
        "{schema}://{template}:{app}@{apikey}/{targets}",
    )

    # Define our template
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            # The App_ID is a UUID
            # such as: 8250eaf6-1a58-489e-b136-7c74a864b434
            "app": {
                "name": _("App ID"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "template": {
                "name": _("Template"),
                "type": "string",
                "private": True,
            },
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_player": {
                "name": _("Target Player ID"),
                "type": "string",
                "map_to": "targets",
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "target_user": {
                "name": _("Target User"),
                "type": "string",
                "prefix": "@",
                "map_to": "targets",
            },
            "target_segment": {
                "name": _("Include Segment"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": True,
                "map_to": "include_image",
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
            "contents": {
                "name": _("Enable Contents"),
                "type": "bool",
                "default": True,
                "map_to": "use_contents",
            },
            "decode": {
                "name": _("Decode Template Args"),
                "type": "bool",
                "default": False,
                "map_to": "decode_tpl_args",
            },
            "template": {
                "alias_of": "template",
            },
            "subtitle": {
                "name": _("Subtitle"),
                "type": "string",
            },
            "language": {
                "name": _("Language"),
                "type": "string",
                "default": "en",
            },
        },
    )

    # Define our token control
    template_kwargs = {
        "custom": {
            "name": _("Custom Data"),
            "prefix": ":",
        },
        "postback": {
            "name": _("Postback Data"),
            "prefix": "+",
        },
    }

    def __init__(
        self,
        app,
        apikey,
        targets=None,
        include_image=True,
        template=None,
        subtitle=None,
        language=None,
        batch=None,
        use_contents=None,
        decode_tpl_args=None,
        custom=None,
        postback=None,
        **kwargs,
    ):
        """Initialize OneSignal."""
        super().__init__(**kwargs)

        # The apikey associated with the account
        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = f"An invalid OneSignal API key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # The App ID associated with the account
        self.app = validate_regex(app)
        if not self.app:
            msg = f"An invalid OneSignal Application ID ({app}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Prepare Batch Mode Flag
        self.batch_size = (
            self.default_batch_size
            if (
                batch
                if batch is not None
                else self.template_args["batch"]["default"]
            )
            else 1
        )

        # Prepare Use Contents Flag
        self.use_contents = bool(
            use_contents
            if use_contents is not None
            else self.template_args["contents"]["default"]
        )

        # Prepare Decode Template Arguments Flag
        self.decode_tpl_args = bool(
            decode_tpl_args
            if decode_tpl_args is not None
            else self.template_args["decode"]["default"]
        )

        # Place a thumbnail image inline with the message body
        self.include_image = include_image

        # Our Assorted Types of Targets
        self.targets = {
            OneSignalCategory.PLAYER: [],
            OneSignalCategory.EMAIL: [],
            OneSignalCategory.USER: [],
            OneSignalCategory.SEGMENT: [],
        }

        # Assign our template (if defined)
        self.template_id = template

        # Assign our subtitle (if defined)
        self.subtitle = subtitle

        # Our Language
        self.language = (
            language.strip().lower()[0:2]
            if language
            else NotifyOneSignal.template_args["language"]["default"]
        )

        if not self.language or len(self.language) != 2:
            msg = f"An invalid OneSignal Language ({language}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Sort our targets
        for _target in parse_list(targets):
            target = _target.strip()
            if len(target) < 2:
                self.logger.debug(f"Ignoring OneSignal Entry: {target}")
                continue

            if target.startswith(
                NotifyOneSignal.template_tokens["target_user"]["prefix"]
            ):

                self.targets[OneSignalCategory.USER].append(target)
                self.logger.debug(
                    "Detected OneSignal UserID:"
                    f" {self.targets[OneSignalCategory.USER][-1]}"
                )
                continue

            if target.startswith(
                NotifyOneSignal.template_tokens["target_segment"]["prefix"]
            ):

                self.targets[OneSignalCategory.SEGMENT].append(target)
                self.logger.debug(
                    "Detected OneSignal Include Segment:"
                    f" {self.targets[OneSignalCategory.SEGMENT][-1]}"
                )
                continue

            result = is_email(target)
            if result:
                self.targets[OneSignalCategory.EMAIL].append(
                    result["full_email"]
                )
                self.logger.debug(
                    "Detected OneSignal Email:"
                    f" {self.targets[OneSignalCategory.EMAIL][-1]}"
                )

            else:
                # Add element as Player ID
                self.targets[OneSignalCategory.PLAYER].append(target)
                self.logger.debug(
                    "Detected OneSignal Player ID:"
                    f" {self.targets[OneSignalCategory.PLAYER][-1]}"
                )

        # Custom Data
        self.custom_data = {}
        if custom and isinstance(custom, dict):
            if self.decode_tpl_args:
                custom = decode_b64_dict(custom)

            self.custom_data.update(custom)

        elif custom:
            msg = (
                "The specified OneSignal Custom Data "
                f"({custom}) are not identified as a dictionary."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Postback Data
        self.postback_data = {}
        if postback and isinstance(postback, dict):
            self.postback_data.update(postback)

        elif postback:
            msg = (
                "The specified OneSignal Postback Data "
                f"({postback}) are not identified as a dictionary."
            )
            self.logger.warning(msg)
            raise TypeError(msg)
        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform OneSignal Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json; charset=utf-8",
            "Authorization": f"Basic {self.apikey}",
        }

        has_error = False
        sent_count = 0

        payload = {
            "app_id": self.app,
            "contents": {
                self.language: body,
            },
            # Sending true wakes your app from background to run custom native
            # code (Apple interprets this as content-available=1).
            # Note: Not applicable if the app is in the "force-quit" state
            #      (i.e app was swiped away). Omit the contents field to
            #      prevent displaying a visible notification.
            "content_available": True,
        }

        if self.template_id:
            # Store template information
            payload["template_id"] = self.template_id

            if not self.use_contents:
                # Only if a template is defined can contents be removed
                del payload["contents"]

        # Set our data if defined
        if self.custom_data:
            payload.update({
                "custom_data": self.custom_data,
            })

        # Set our postback data if defined
        if self.postback_data:
            payload.update({
                "data": self.postback_data,
            })

        if title:
            # Display our title if defined
            payload.update(
                {
                    "headings": {
                        self.language: title,
                    }
                }
            )

        if self.subtitle:
            payload.update({
                "subtitle": {
                    self.language: self.subtitle,
                },
            })

        # Acquire our large_icon image URL (if set)
        image_url = (
            None if not self.include_image else self.image_url(notify_type)
        )
        if image_url:
            payload["large_icon"] = image_url

        # Acquire our small_icon image URL (if set)
        image_url = (
            None
            if not self.include_image
            else self.image_url(notify_type, image_size=NotifyImageSize.XY_32)
        )
        if image_url:
            payload["small_icon"] = image_url

        for category in ONESIGNAL_CATEGORIES:
            # Create a pointer to our list of targets for specified category
            targets = self.targets[category]
            for index in range(0, len(targets), self.batch_size):
                payload[category] = targets[index : index + self.batch_size]

                # Track our sent count
                sent_count += len(payload[category])

                self.logger.debug(
                    "OneSignal POST URL:"
                    f" {self.notify_url} "
                    f"(cert_verify={self.verify_certificate!r})"
                )
                self.logger.debug(f"OneSignal Payload: {payload!s}")

                # Always call throttle before any remote server i/o is made
                self.throttle()
                try:
                    r = requests.post(
                        self.notify_url,
                        data=dumps(payload),
                        headers=headers,
                        verify=self.verify_certificate,
                        timeout=self.request_timeout,
                    )
                    if r.status_code not in (
                        requests.codes.ok,
                        requests.codes.no_content,
                    ):
                        # We had a problem
                        status_str = NotifyOneSignal.http_response_code_lookup(
                            r.status_code
                        )

                        self.logger.warning(
                            "Failed to send OneSignal notification: "
                            "{}{}error={}.".format(
                                status_str,
                                ", " if status_str else "",
                                r.status_code,
                            )
                        )

                        self.logger.debug("Response Details:\r\n%s", r.content)

                        has_error = True

                    else:
                        self.logger.info("Sent OneSignal notification.")

                except requests.RequestException as e:
                    self.logger.warning(
                        "A Connection error occurred sending OneSignal "
                        "notification."
                    )
                    self.logger.debug("Socket Exception: %s", str(e))

                    has_error = True

        if not sent_count:
            # There is no one to notify; we need to capture this and not
            # return a valid
            self.logger.warning("There are no OneSignal targets to notify")
            return False

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.template_id,
            self.app,
            self.apikey,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
            "batch": "yes" if self.batch_size > 1 else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        custom_data, needs_decoding = encode_b64_dict(self.custom_data)
        # custom_data, needs_decoding = self.custom_data, False
        # Save our template data
        params.update({f":{k}": v for k, v in custom_data.items()})

        # Save our postback data
        params.update({f"+{k}": v for k, v in self.postback_data.items()})

        if self.use_contents != self.template_args["contents"]["default"]:
            params["contents"] = "yes" if self.use_contents else "no"

        if (
            self.decode_tpl_args != self.template_args["decode"]["default"]
            or needs_decoding
        ):
            params["decode"] = (
                "yes" if (self.decode_tpl_args or needs_decoding) else "no"
            )

        return "{schema}://{tp_id}{app}@{apikey}/{targets}?{params}".format(
            schema=self.secure_protocol,
            tp_id=(
                "{}:".format(self.pprint(self.template_id, privacy, safe=""))
                if self.template_id
                else ""
            ),
            app=self.pprint(self.app, privacy, safe=""),
            apikey=self.pprint(self.apikey, privacy, safe=""),
            targets="/".join(
                chain(
                    [
                        NotifyOneSignal.quote(x)
                        for x in self.targets[OneSignalCategory.PLAYER]
                    ],
                    [
                        NotifyOneSignal.quote(x)
                        for x in self.targets[OneSignalCategory.EMAIL]
                    ],
                    [
                        NotifyOneSignal.quote(
                            "{}{}".format(
                                NotifyOneSignal.template_tokens["target_user"][
                                    "prefix"
                                ],
                                x,
                            ),
                            safe="",
                        )
                        for x in self.targets[OneSignalCategory.USER]
                    ],
                    [
                        NotifyOneSignal.quote(
                            "{}{}".format(
                                NotifyOneSignal.template_tokens[
                                    "target_segment"
                                ]["prefix"],
                                x,
                            ),
                            safe="",
                        )
                        for x in self.targets[OneSignalCategory.SEGMENT]
                    ],
                )
            ),
            params=NotifyOneSignal.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        #
        # Factor batch into calculation
        #
        if self.batch_size > 1:
            # Batches can only be sent by group (you can't combine groups into
            # a single batch)
            total_targets = 0
            for _k, m in self.targets.items():
                targets = len(m)
                total_targets += int(targets / self.batch_size) + (
                    1 if targets % self.batch_size else 0
                )
            return total_targets

        # Normal batch count; just count the targets
        return sum([len(m) for _, m in self.targets.items()])

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        if not results.get("password"):
            # The APP ID identifier associated with the account
            results["app"] = NotifyOneSignal.unquote(results["user"])

        else:
            # The APP ID identifier associated with the account
            results["app"] = NotifyOneSignal.unquote(results["password"])
            # The Template ID
            results["template"] = NotifyOneSignal.unquote(results["user"])

        # Get Image Boolean (if set)
        results["include_image"] = parse_bool(
            results["qsd"].get(
                "image", NotifyOneSignal.template_args["image"]["default"]
            )
        )

        # Get Batch Boolean (if set)
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch", NotifyOneSignal.template_args["batch"]["default"]
            )
        )

        # Get Use Contents Boolean (if set)
        results["use_contents"] = parse_bool(
            results["qsd"].get(
                "contents",
                NotifyOneSignal.template_args["contents"]["default"],
            )
        )

        # Get Use Contents Boolean (if set)
        results["decode_tpl_args"] = parse_bool(
            results["qsd"].get(
                "decode", NotifyOneSignal.template_args["decode"]["default"]
            )
        )

        # The API Key is stored in the hostname
        results["apikey"] = NotifyOneSignal.unquote(results["host"])

        # Get our Targets
        results["targets"] = NotifyOneSignal.split_path(results["fullpath"])

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyOneSignal.parse_list(
                results["qsd"]["to"]
            )

        if "app" in results["qsd"] and len(results["qsd"]["app"]):
            results["app"] = NotifyOneSignal.unquote(results["qsd"]["app"])

        if "apikey" in results["qsd"] and len(results["qsd"]["apikey"]):
            results["apikey"] = NotifyOneSignal.unquote(
                results["qsd"]["apikey"]
            )

        if "template" in results["qsd"] and len(results["qsd"]["template"]):
            results["template"] = NotifyOneSignal.unquote(
                results["qsd"]["template"]
            )

        if "subtitle" in results["qsd"] and len(results["qsd"]["subtitle"]):
            results["subtitle"] = NotifyOneSignal.unquote(
                results["qsd"]["subtitle"]
            )

        if "lang" in results["qsd"] and len(results["qsd"]["lang"]):
            results["language"] = NotifyOneSignal.unquote(
                results["qsd"]["lang"]
            )

        # Store our custom data
        results["custom"] = results["qsd:"]

        # Store our postback data
        results["postback"] = results["qsd+"]

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this plugin, sign up with Hampager (you need to be a licensed
# ham radio operator
#  http://www.hampager.de/
#
# You're done at this point, you only need to know your user/pass that
# you signed up with.

#  The following URLs would be accepted by Apprise:
#   - dapnet://{user}:{password}@{callsign}
#   - dapnet://{user}:{password}@{callsign1}/{callsign2}

# Optional parameters:
#   - priority (NORMAL or EMERGENCY). Default: NORMAL
#   - txgroups --> comma-separated list of DAPNET transmitter
#                           groups. Default: 'dl-all'
#                           https://hampager.de/#/transmitters/groups

from json import dumps

# The API reference used to build this plugin was documented here:
#  https://hampager.de/dokuwiki/doku.php#dapnet_api
#
import requests
from requests.auth import HTTPBasicAuth

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_call_sign, parse_bool, parse_call_sign, parse_list
from .base import NotifyBase


class DapnetPriority:
    NORMAL = 0
    EMERGENCY = 1


DAPNET_PRIORITIES = {
    DapnetPriority.NORMAL: "normal",
    DapnetPriority.EMERGENCY: "emergency",
}


DAPNET_PRIORITY_MAP = {
    # Maps against string 'normal'
    "n": DapnetPriority.NORMAL,
    # Maps against string 'emergency'
    "e": DapnetPriority.EMERGENCY,
    # Entries to additionally support (so more like Dapnet's API)
    "0": DapnetPriority.NORMAL,
    "1": DapnetPriority.EMERGENCY,
}


class NotifyDapnet(NotifyBase):
    """A wrapper for DAPNET / Hampager Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Dapnet"

    # The services URL
    service_url = "https://hampager.de/"

    # The default secure protocol
    secure_protocol = "dapnet"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_dapnet"

    # Dapnet uses the http protocol with JSON requests
    notify_url = "http://www.hampager.de:8080/calls"

    # The maximum length of the body
    body_maxlen = 80

    # A title can not be used for Dapnet Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # The maximum amount of emails that can reside within a single transmission
    default_batch_size = 50

    # Define object templates
    templates = ("{schema}://{user}:{password}@{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user": {
                "name": _("User Name"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_callsign": {
                "name": _("Target Callsign"),
                "type": "string",
                "regex": (
                    r"^[a-z0-9]{2,5}(-[a-z0-9]{1,2})?$",
                    "i",
                ),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "name": _("Target Callsign"),
                "type": "string",
                "map_to": "targets",
            },
            "priority": {
                "name": _("Priority"),
                "type": "choice:int",
                "values": DAPNET_PRIORITIES,
                "default": DapnetPriority.NORMAL,
            },
            "txgroups": {
                "name": _("Transmitter Groups"),
                "type": "string",
                "default": "dl-all",
                "private": True,
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
        },
    )

    def __init__(
        self, targets=None, priority=None, txgroups=None, batch=False, **kwargs
    ):
        """Initialize Dapnet Object."""
        super().__init__(**kwargs)

        # Parse our targets
        self.targets = []

        # The Priority of the message
        self.priority = int(
            NotifyDapnet.template_args["priority"]["default"]
            if priority is None
            else next(
                (
                    v
                    for k, v in DAPNET_PRIORITY_MAP.items()
                    if str(priority).lower().startswith(k)
                ),
                NotifyDapnet.template_args["priority"]["default"],
            )
        )

        if not (self.user and self.password):
            msg = "A Dapnet user/pass was not provided."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Get the transmitter group
        self.txgroups = parse_list(
            txgroups
            if txgroups
            else NotifyDapnet.template_args["txgroups"]["default"]
        )

        # Prepare Batch Mode Flag
        self.batch = batch

        for target in parse_call_sign(targets):
            # Validate targets and drop bad ones:
            result = is_call_sign(target)
            if not result:
                self.logger.warning(
                    f"Dropping invalid Amateur radio call sign ({target}).",
                )
                continue

            # Store callsign without SSID and ignore duplicates
            if result["callsign"] not in self.targets:
                self.targets.append(result["callsign"])

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Dapnet Notification."""

        if not self.targets:
            # There is no one to email; we're done
            self.logger.warning(
                "There are no Amateur radio callsigns to notify"
            )
            return False

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json; charset=utf-8",
        }

        # error tracking (used for function return)
        has_error = False

        # Create a copy of the targets list
        targets = list(self.targets)

        for index in range(0, len(targets), batch_size):

            # prepare JSON payload
            payload = {
                "text": body,
                "callSignNames": targets[index : index + batch_size],
                "transmitterGroupNames": self.txgroups,
                "emergency": self.priority == DapnetPriority.EMERGENCY,
            }

            self.logger.debug(f"DAPNET POST URL: {self.notify_url}")
            self.logger.debug(f"DAPNET Payload: {dumps(payload)}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=dumps(payload),
                    headers=headers,
                    auth=HTTPBasicAuth(
                        username=self.user, password=self.password
                    ),
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code != requests.codes.created:
                    # We had a problem

                    self.logger.warning(
                        "Failed to send DAPNET notification {} to {}: "
                        "error={}.".format(
                            payload["text"],
                            f" to {self.targets}",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True

                else:
                    self.logger.info(
                        "Sent '{}' DAPNET notification {}".format(
                            payload["text"], f"to {self.targets}"
                        )
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending DAPNET "
                    f"notification to {self.targets}"
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True

        return not has_error

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "priority": (
                DAPNET_PRIORITIES[self.template_args["priority"]["default"]]
                if self.priority not in DAPNET_PRIORITIES
                else DAPNET_PRIORITIES[self.priority]
            ),
            "batch": "yes" if self.batch else "no",
            "txgroups": ",".join(self.txgroups),
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Setup Authentication
        auth = "{user}:{password}@".format(
            user=NotifyDapnet.quote(self.user, safe=""),
            password=self.pprint(
                self.password, privacy, mode=PrivacyMode.Secret, safe=""
            ),
        )

        return "{schema}://{auth}{targets}?{params}".format(
            schema=self.secure_protocol,
            auth=auth,
            targets="/".join(
                [self.pprint(x, privacy, safe="") for x in self.targets]
            ),
            params=NotifyDapnet.urlencode(params),
        )

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.user, self.password)

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        #
        # Factor batch into calculation
        #
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if batch_size > 1:
            targets = int(targets / batch_size) + (
                1 if targets % batch_size else 0
            )

        return targets

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # All elements are targets
        results["targets"] = [NotifyDapnet.unquote(results["host"])]

        # All entries after the hostname are additional targets
        results["targets"].extend(NotifyDapnet.split_path(results["fullpath"]))

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyDapnet.parse_list(results["qsd"]["to"])

        # Set our priority
        if "priority" in results["qsd"] and len(results["qsd"]["priority"]):
            results["priority"] = NotifyDapnet.unquote(
                results["qsd"]["priority"]
            )

        # Check for one or multiple transmitter groups (comma separated)
        # and split them up, when necessary
        if "txgroups" in results["qsd"]:
            results["txgroups"] = [
                x.lower()
                for x in NotifyDapnet.parse_list(results["qsd"]["txgroups"])
            ]

        # Get Batch Mode Flag
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch", NotifyDapnet.template_args["batch"]["default"]
            )
        )

        return results
